#include <stdio.h>
#include "ga-cfp.h"

/*=====================================================*/
/* Genetic Algorithm.                                  */
/* Elitist; best chromosome is always in Population[0] */
/* Two-point crossover is used.
/*=====================================================*/

extern double drand48();
extern int verbosity;
extern int No_Features;
extern int linear[];
extern INSTANCE trainData[];
extern int No_Trainers;
extern FILE *logf;

int No_Chromosomes, No_Generations;
float pc, pm;
float Population[MAX_CHROMOSOMES][MAX_GENES],
      Old_Population[MAX_CHROMOSOMES][MAX_GENES];
float Fitness[MAX_CHROMOSOMES],
      Old_Fitness[MAX_CHROMOSOMES],
      Wheel[MAX_CHROMOSOMES];

int No_Genes;

float average (Array, size)
float *Array;
int size;
{
  int i;
  float sum = 0.0;

  for (i=0; i < size; i++)
    sum += Array[i];
  return(sum/size);
}
  
void set_ga_defaults()
{
   pc = 0.7;
   pm = 0.01;
   No_Chromosomes = 100;
   No_Generations = 50;
 }

float random_value(max)
float max;
{
  return((float)drand48() * max);
}

void initialize_ga ()
{
  int c, f, g, i;
  float max[MAX_FEATURES], min[MAX_FEATURES], range[MAX_FEATURES];

  for (f=0; f < No_Features; f++) { /* determine ranges */
    max[f] = -INFINITE; min[f] = INFINITE;
    for (i=0; i < No_Trainers; i++) 
      if (trainData[i].status[f] == KNOWN) {
	if (trainData[i].value[f] > max[f]) 
	  max[f] = trainData[i].value[f];
	if (trainData[i].value[f] < min[f]) 
	  min[f] = trainData[i].value[f];
      }
    range[f] = max[f] - min[f];
  }

  if (verbosity > 0) {
    fprintf (logf, "Ranges:\n");
    for (f=0; f < No_Features; f++)
      fprintf (logf, "R[%d]=%g ", f, range[f]);
    fprintf (logf, "\n");
  }

  srand48((long)17); /* initialize the random number generator */

  for (c=0; c < No_Chromosomes; c++) {
    Population[c][0] = random_value (0.1); /* delta */
    for (g=1; g < No_Genes; g++)           /* generalization limit */
      if (linear[g-1])
	Population[c][g] = range[g-1] * 0.1 * (0.5 + random_value(1.0));
      else
	Population[c][g] = 0.0;  /*gen_limit = 0 for nominal features */
    Fitness[c] = 0.0;
  }
}

void evaluate ()
{
  int c;

  for (c=1; c < No_Chromosomes; c++) 
    if (Fitness[c] == 0.0) 
      Fitness[c] = fitness(Population[c]);
}

void elitism()
{
  int g, c, best;

  best = 0;
  for (c=1; c < No_Chromosomes; c++) 
    if (Fitness[best] <= Fitness[c]) 
      best = c;
  /* copy best to Poulation[0] */
  for (g=0; g < No_Genes; g++)
    Population[0][g] = Population[best][g];
  Fitness[0] = Fitness[best];
}

void reproduce ()
{
   int g, c, selection;
   float pointer, sum_fitness;

   /* copy Population to Old_Population */
   for (c=0; c < No_Chromosomes; c++) {
     for (g=0; g < No_Genes; g++)
       Old_Population[c][g] = Population[c][g];
     Old_Fitness[c] = Fitness[c];
     }
  
   /* set the Wheel */   
   sum_fitness = 0.0;
   for (c=0; c < No_Chromosomes; c++) {
     sum_fitness += Fitness[c];
     Wheel[c] = sum_fitness;
   }
   /* best chromosome is already in Population[0] */
   for (c=1; c < No_Chromosomes; c++) {
     pointer = random_value (sum_fitness);
     for (selection=0; pointer > Wheel[selection]; selection++);
     for (g=0; g < No_Genes; g++)
       Population[c][g] = Old_Population[selection][g];
     Fitness[c] = Old_Fitness[selection];
   }
}

void crossover ()  /* uniform crossover */
{
  int c, g;
  float tmp;

  for (c=1; c < No_Chromosomes-2; c += 2) 
    if (random_value(1.0) < pc) {
      /* apply crossover */
      Fitness[c] = Fitness[c+1] = 0.0; /* reevaluate */

      for (g=0; g < No_Genes; g++) 
	if (random_value(1.0) < 0.5) { /* swap */
	  tmp = Population[c][g];
	  Population[c][g] = Population[c+1][g];
	  Population[c+1][g] = tmp;
	}
    }
}

void mutation ()
{
  int c, g, modified;

  for (c=1; c < No_Chromosomes; c++) {
    modified = 0;
    for (g=0; g < No_Genes; g++) 
      if (random_value(1.0) < pm) {
	Population[c][g] *= (0.5 + random_value(1.0));   /* +- 50% */
	modified = 1;
      }
    if (modified)
      Fitness[c] =0.0;
  }  
}

void ga_search()
{
  int g;

  initialize_ga();
  evaluate();
  elitism();
  for (g=0; g < No_Generations; g++) {
    reproduce();
    crossover();
    mutation();
    evaluate();
    elitism();
    if (verbosity > 0)
      show_population(); 
    printf("Gen: %d, Best: %g, average: %g\n", 
	   g, Fitness[0], average(Fitness, No_Chromosomes));
  }
}

float ga_learn_parameters (delta, gen_limit)
float *delta, *gen_limit;
{
  int g;

  No_Genes = No_Features + 1;

  ga_search();

  /* set the learned parameters back */
  *delta = Population[0][0];
  for (g=1; g < No_Genes; g++)
    gen_limit[g-1] = Population[0][g];

  return(Fitness[0]);
}
